#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0

"""
Change human readable data access monitoring-based operation schemes to the low
level input for the '<debugfs>/damon/schemes' file.  Below is an example of the
schemes written in the human readable format:

    # format is:
    # <min/max size> <min/max frequency (0-100)> <min/max age> <action>
    #
    # B/K/M/G/T for Bytes/KiB/MiB/GiB/TiB
    # us/ms/s/m/h/d for micro-seconds/milli-seconds/seconds/minutes/hours/days
    # 'min/max' for possible min/max value.

    # if a region keeps a high access frequency for >=100ms, put the region on
    # the head of the LRU list (call madvise() with MADV_WILLNEED).
    min    max      80      max     100ms   max willneed

    # if a region keeps a low access frequency for >=200ms and <=one hour, put
    # the region on the tail of the LRU list (call madvise() with MADV_COLD).
    min     max     10      20      200ms   1h  cold

    # if a region keeps a very low access frequency for >=60 seconds, swap out
    # the region immediately (call madvise() with MADV_PAGEOUT).
    min     max     0       10      60s     max pageout

    # if a region of a size >=2MiB keeps a very high access frequency for
    # >=100ms, let the region to use huge pages (call madvise() with
    # MADV_HUGEPAGE).
    2M      max     90      100     100ms   max hugepage

    # If a regions of a size >=2MiB keeps small access frequency for >=100ms,
    # avoid the region using huge pages (call madvise() with MADV_NOHUGEPAGE).
    2M      max     0       25      100ms   max nohugepage
"""

import argparse
import os
import platform

import _damon

uint_max = 2**32 - 1
ulong_max = 2**64 - 1
if platform.architecture()[0] != '64bit':
    ulong_max = 2**32 - 1

unit_to_bytes = {'B': 1, 'K': 1024, 'M': 1024 * 1024, 'G': 1024 * 1024 * 1024,
        'T': 1024 * 1024 * 1024 * 1024}

def text_to_bytes(txt):
    if txt == 'min':
        return 0
    if txt == 'max':
        return ulong_max

    unit = txt[-1]
    number = float(txt[:-1])
    return int(number * unit_to_bytes[unit])

unit_to_usecs = {'us': 1, 'ms': 1000, 's': 1000 * 1000, 'm': 60 * 1000 * 1000,
        'h': 60 * 60 * 1000 * 1000, 'd': 24 * 60 * 60 * 1000 * 1000}

def text_to_us(txt):
    if txt == 'min':
        return 0
    if txt == 'max':
        return uint_max

    unit = txt[-2:]
    if unit in ['us', 'ms']:
        number = float(txt[:-2])
    else:
        unit = txt[-1]
        number = float(txt[:-1])
    return number * unit_to_usecs[unit]

def text_to_ms(txt):
    return int(text_to_us(txt) / 1000)

damos_action_to_int = {'willneed': 0, 'cold': 1, 'pageout': 2, 'hugepage': 3,
        'nohugepage': 4, 'stat': 5, 'lru_prio': 6, 'lru_deprio': 7}

def text_to_nr_accesses(txt, max_nr_accesses):
    if txt == 'min':
        return 0
    if txt == 'max':
        return max_nr_accesses

    return int(float(txt) * max_nr_accesses / 100)

def text_percent_to_nr_accesses_permil(txt):
    if txt == 'min':
        return 0
    if txt == 'max':
        return 1000
    return float(txt) * 10

damos_wmark_metric_to_int = {'none': 0, 'free_mem_rate': 1}

def text_to_damos_wmark_metric(txt):
    return damos_wmark_metric_to_int[txt.lower()]

def damo_scheme_to_damos(line, scheme_version):
    fields = line.split()
    expected_lengths = [7, 9, 12, 17, 18]
    if not len(fields) in expected_lengths:
        print('expected %s fields, but \'%s\'' % (expected_lengths, line))
        exit(1)

    try:
        min_sz = text_to_bytes(fields[0])
        max_sz = text_to_bytes(fields[1])
        min_nr_accesses = text_percent_to_nr_accesses_permil(fields[2])
        max_nr_accesses = text_percent_to_nr_accesses_permil(fields[3])
        min_age_us = text_to_us(fields[4])
        max_age_us = text_to_us(fields[5])
        action_txt = fields[6].lower()
        quota_ms = 0
        quota_sz = 0
        window_ms = ulong_max
        weight_sz = 0
        weight_nr_accesses = 0
        weight_age = 0
        wmarks_txt = 'none'
        wmarks_metric = text_to_damos_wmark_metric('none')
        wmarks_interval = 0
        wmarks_high = 0
        wmarks_mid = 0
        wmarks_low = 0
        if len(fields) <= 17:
            if len(fields) >= 9:
                quota_sz = text_to_bytes(fields[7])
                window_ms = text_to_ms(fields[8])
            if len(fields) >= 12:
                weight_sz = int(fields[9])
                weight_nr_accesses = int(fields[10])
                weight_age = int(fields[11])
            if len(fields) == 17:
                wmarks_txt = fields[12].lower()
                wmarks_metric = text_to_damos_wmark_metric(fields[12])
                wmarks_interval = text_to_us(fields[13])
                wmarks_high = int(fields[14])
                wmarks_mid = int(fields[15])
                wmarks_low = int(fields[16])
        elif len(fields) == 18:
            quota_ms = text_to_ms(fields[7])
            quota_sz = text_to_bytes(fields[8])
            window_ms = text_to_ms(fields[9])
            weight_sz = int(fields[10])
            weight_nr_accesses = int(fields[11])
            weight_age = int(fields[12])
            wmarks_txt = fields[13].lower()
            wmarks_metric = text_to_damos_wmark_metric(fields[13])
            wmarks_interval = text_to_us(fields[14])
            wmarks_high = int(fields[15])
            wmarks_mid = int(fields[16])
            wmarks_low = int(fields[17])

    except:
        print('wrong input field')
        raise

    return _damon.Damos(_damon.DamosAccessPattern(min_sz, max_sz,
        min_nr_accesses, max_nr_accesses, min_age_us, max_age_us),
        action_txt,
        _damon.DamosQuota(quota_ms, quota_sz, window_ms, weight_sz,
            weight_nr_accesses, weight_age),
        _damon.DamosWatermarks(wmarks_txt, wmarks_interval, wmarks_high,
            wmarks_mid, wmarks_low))

def damos_to_debugfs_input(damos, sample_interval, aggr_interval,
        scheme_version):
    pattern = damos.access_pattern
    quotas = damos.quotas
    watermarks = damos.watermarks

    max_nr_accesses = aggr_interval / sample_interval
    v0_scheme = '%d\t%d\t%d\t%d\t%d\t%d\t%d' % (
            pattern.min_sz_bytes, pattern.max_sz_bytes,
            int(pattern.min_nr_accesses_permil * max_nr_accesses / 1000),
            int(pattern.max_nr_accesses_permil * max_nr_accesses / 1000),
            pattern.min_age_us / aggr_interval,
            pattern.max_age_us / aggr_interval,
            damos_action_to_int[damos.action])
    v1_scheme = '%s\t%d\t%d' % (v0_scheme,
            quotas.sz_bytes, quotas.reset_interval_ms)
    v2_scheme = '%s\t%d\t%d\t%d' % (v1_scheme,
            quotas.weight_sz_permil, quotas.weight_nr_accesses_permil,
            quotas.weight_age_permil)
    v3_scheme = '%s\t%d\t%d\t%d\t%d\t%d' % (v2_scheme,
            text_to_damos_wmark_metric(watermarks.metric),
            watermarks.interval_us, watermarks.high_permil,
            watermarks.mid_permil, watermarks.low_permil)
    v4_scheme = '%s\t' % v0_scheme + '\t'.join('%d' % x for x in [quotas.time_ms,
        quotas.sz_bytes, quotas.reset_interval_ms, quotas.weight_sz_permil,
        quotas.weight_nr_accesses_permil, quotas.weight_age_permil,
        text_to_damos_wmark_metric(watermarks.metric), watermarks.interval_us,
        watermarks.high_permil, watermarks.mid_permil, watermarks.low_permil])

    if scheme_version == 0:
        return v0_scheme
    elif scheme_version == 1:
        return v1_scheme
    elif scheme_version == 2:
        return v2_scheme
    elif scheme_version == 3:
        return v3_scheme
    elif scheme_version == 4:
        return v4_scheme
    else:
        print('Unsupported scheme version: %d' % scheme_version)
        exit(1)

def damo_schemes_split_remove_comments(schemes):
    raw_lines = schemes.split('\n')
    clean_lines = []
    for line in raw_lines:
        if line.startswith('#'):
            continue
        line = line.strip()
        if line == '':
            continue
        clean_lines.append(line)
    return clean_lines

def convert(schemes, target, sample_interval, aggr_interval, scheme_version):
    if os.path.isfile(schemes):
        with open(schemes, 'r') as f:
            schemes = f.read()

    damos_list = []
    debugfs_schemes_input_lines = []
    for line in damo_schemes_split_remove_comments(schemes):
        damos = damo_scheme_to_damos(line, scheme_version)
        damos_list.append(damos)
        debugfs_schemes_input_lines.append(damos_to_debugfs_input(damos,
            sample_interval, aggr_interval, scheme_version))
    if target == 'debugfs schemes input':
        return '\n'.join(debugfs_schemes_input_lines)
    elif target == 'damos':
        return damos_list
    else:
        print('_convert_damos.convert() received unsupported target \'%s\'',
                target)

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('input', metavar='<file or schemes in text>',
            help='input file describing the schemes or the schemes')
    parser.add_argument('-s', '--sample', metavar='<interval>', type=int,
            default=5000, help='sampling interval (us)')
    parser.add_argument('-a', '--aggr', metavar='<interval>', type=int,
            default=100000, help='aggregation interval (us)')
    parser.add_argument('--scheme_version', metavar='<version>', type=int,
            choices=range(0, 5), default=4, help='destination scheme version')
    args = parser.parse_args()

    sample_interval = args.sample
    aggr_interval = args.aggr
    scheme_ver = args.scheme_version

    print(convert(args.input, 'debugfs schemes input', sample_interval,
        aggr_interval, scheme_ver))

if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0

"""
Contains core functions for DAMON debugfs control.
"""

import os
import subprocess

import _damon

debugfs = '/sys/kernel/debug'
debugfs_damon = os.path.join(debugfs, 'damon')
debugfs_version = os.path.join(debugfs_damon, 'version')
debugfs_attrs = os.path.join(debugfs_damon, 'attrs')
debugfs_record = os.path.join(debugfs_damon, 'record')
debugfs_schemes = os.path.join(debugfs_damon, 'schemes')
debugfs_target_ids = os.path.join(debugfs_damon, 'target_ids')
debugfs_init_regions = os.path.join(debugfs_damon, 'init_regions')
debugfs_monitor_on = os.path.join(debugfs_damon, 'monitor_on')

def set_target_id(tid):
    try:
        with open(debugfs_target_ids, 'w') as f:
            f.write('%s\n' % tid)
    except Exception as e:
        return e

def set_target(tid, init_regions=[]):
    rc = set_target_id(tid)
    if rc:
        return rc

    if not debugfs_init_regions:
        return 0

    if feature_supported('init_regions_target_idx'):
        tid = 0
    elif tid == 'paddr':
        tid = 42

    string = ' '.join(['%s %d %d' % (tid, r[0], r[1]) for r in init_regions])
    return subprocess.call('echo "%s" > %s' % (string, debugfs_init_regions),
            shell=True, executable='/bin/bash')

def turn_damon(on_off):
    return subprocess.call('echo %s > %s' % (on_off, debugfs_monitor_on),
            shell=True, executable='/bin/bash')

def is_damon_running():
    with open(debugfs_monitor_on, 'r') as f:
        return f.read().strip() == 'on'

def current_attrs():
    with open(debugfs_attrs, 'r') as f:
        attrs = f.read().split()
    attrs = [int(x) for x in attrs]

    if debugfs_record:
        with open(debugfs_record, 'r') as f:
            rattrs = f.read().split()
        attrs.append(int(rattrs[0]))
        attrs.append(rattrs[1])
    else:
        attrs += [None, None]

    if debugfs_schemes:
        with open(debugfs_schemes, 'r') as f:
            schemes = f.read()
        # The last two fields in each line are statistics.  Remove those.
        schemes = [' '.join(x.split()[:-2]) for x in schemes.strip().split('\n')]
        attrs.append('\n'.join(schemes))
    else:
        attrs.append(None)

    return _damon.Attrs(*attrs)

feature_supports = None

def feature_supported(feature):
    if feature_supports == None:
        initialize()

    return feature_supports[feature]

def get_supported_features():
    if feature_supports == None:
        initialize()
    return feature_supports

def test_debugfs_file(path, input_str, expected):
    passed = False
    with open(path, 'r') as f:
        orig_value = f.read()
        if orig_value == '':
            orig_value = '\n'
    if os.path.basename(path) == 'target_ids' and orig_value == '42\n':
        orig_value = 'paddr\n'
    try:
        with open(path, 'w') as f:
            f.write(input_str)
        with open(path, 'r') as f:
            if f.read() == expected:
                passed = True
    except Exception as e:
        passed = False
    with open(path, 'w') as f:
        f.write(orig_value)
    return passed

def test_debugfs_file_schemes(nr_fields):
    input_str = ' '.join(['1'] * nr_fields)
    expected = '%s 0 0\n' % input_str

    return test_debugfs_file(debugfs_schemes, input_str, expected)

def test_debugfs_file_schemes_stat_extended(nr_fields):
    input_str = ' '.join(['1'] * nr_fields)
    expected = '%s 0 0 0 0 0\n' % input_str

    return test_debugfs_file(debugfs_schemes, input_str, expected)

def test_init_regions_version(paddr_supported):
    # Save previous values
    with open(debugfs_target_ids, 'r') as f:
        orig_target_ids = f.read()
        if orig_target_ids == '':
            orig_target_ids = '\n'
        if orig_target_ids == '42\n':
            orig_target_ids = 'paddr\n'
    with open(debugfs_init_regions, 'r') as f:
        orig_init_regions = f.read()
        if orig_init_regions == '':
            orig_init_regions = '\n'

    # Test
    with open(debugfs_target_ids, 'w') as f:
        if paddr_supported:
            f.write('paddr\n')
        else:
            f.write('%d\n' % os.getpid())

    if paddr_supported:
        v1_input = '42 100 200'
    else:
        v1_input = '%d 100 200' % os.getpid()
    try:
        with open(debugfs_init_regions, 'w') as f:
            f.write(v1_input)
    except IOError as e:
        version = 2
    with open(debugfs_init_regions, 'r') as f:
        if f.read().strip() == v1_input:
            version = 1
        else:
            version = 2

    # Restore previous values
    try:
        with open(debugfs_target_ids, 'w') as f:
            f.write(orig_target_ids)
        with open(debugfs_init_regions, 'w') as f:
            f.write(orig_init_regions)
    except IOError:
        # Previous value might be invalid now (e.g., process terminated)
        pass
    return version

def update_supported_features():
    global feature_supports
    if feature_supports != None:
        return None
    feature_supports = {x: False for x in _damon.features}

    err = set_damon_debugfs_paths()
    if err != None:
        return err

    if debugfs_record != None:
        feature_supports['record'] = True
    if debugfs_schemes != None:
        feature_supports['schemes'] = True

    # virtual address space has supported since the beginning
    feature_supports['vaddr'] = True
    if test_debugfs_file(debugfs_target_ids, 'paddr\n', '42\n'):
        feature_supports['paddr'] = True

    if debugfs_init_regions != None:
        feature_supports['init_regions'] = True
        init_regions_version = test_init_regions_version(
                feature_supports['paddr'])
        if init_regions_version == 2:
            feature_supports['init_regions_target_idx'] = True

    if debugfs_schemes != None:
        if test_debugfs_file_schemes(9):
            feature_supports['schemes_speed_limit'] = True
        elif test_debugfs_file_schemes(12):
            feature_supports['schemes_speed_limit'] = True
            feature_supports['schemes_prioritization'] = True
        elif test_debugfs_file_schemes(17):
            feature_supports['schemes_speed_limit'] = True
            feature_supports['schemes_prioritization'] = True
            feature_supports['schemes_wmarks'] = True
        elif test_debugfs_file_schemes(18):
            feature_supports['schemes_speed_limit'] = True
            feature_supports['schemes_prioritization'] = True
            feature_supports['schemes_wmarks'] = True
            feature_supports['schemes_quotas'] = True
        elif test_debugfs_file_schemes_stat_extended(18):
            feature_supports['schemes_speed_limit'] = True
            feature_supports['schemes_prioritization'] = True
            feature_supports['schemes_wmarks'] = True
            feature_supports['schemes_quotas'] = True
            feature_supports['schemes_stat_succ'] = True
            feature_supports['schemes_stat_qt_exceed'] = True
    return None

def set_damon_debugfs_paths():
    'Set global variables for DAMON debugfs path.  Return error if failed'
    global debugfs_damon
    global debugfs_version
    global debugfs_attrs
    global debugfs_record
    global debugfs_schemes
    global debugfs_target_ids
    global debugfs_init_regions
    global debugfs_monitor_on

    if not os.path.isdir(debugfs_damon):
        return 'damon debugfs dir (%s) not found' % debugfs_damon

    for f in [debugfs_version, debugfs_attrs, debugfs_record, debugfs_schemes,
            debugfs_target_ids, debugfs_init_regions, debugfs_monitor_on]:
        # f could be None if this function is called before
        if f == None:
            continue
        if not os.path.isfile(f):
            if f == debugfs_version:
                debugfs_version = None
            elif f == debugfs_record:
                debugfs_record = None
            elif f == debugfs_schemes:
                debugfs_schemes = None
            elif f == debugfs_init_regions:
                debugfs_init_regions = None
            else:
                return 'damon debugfs file (%s) not found' % f
    return None

def initialize(args, skip_dirs_population=False):
    err = update_supported_features()
    if err:
        return err
    return None

def attr_str(attrs):
    return '%s %s %s %s %s ' % (attrs.sample_interval, attrs.aggr_interval,
            attrs.regions_update_interval, attrs.min_nr_regions,
            attrs.max_nr_regions)

def record_str(attrs):
    return '%s %s ' % (attrs.rbuf_len, attrs.rfile_path)

def attrs_apply(attrs):
    for scheme in attrs.schemes.split('\n'):
        if scheme.strip() == '':
            continue
        action = int(scheme.split()[6])
        if action > 5:
            print('dbgfs unsupported DAMOS action')
            exit(1)
    ret = subprocess.call('echo %s > %s' % (attr_str(attrs), debugfs_attrs),
            shell=True, executable='/bin/bash')
    if ret:
        return ret
    if debugfs_record:
        ret = subprocess.call('echo %s > %s' % (record_str(attrs),
            debugfs_record), shell=True, executable='/bin/bash')
        if ret:
            return ret
    if not debugfs_schemes:
        return 0
    return subprocess.call('echo %s > %s' % (
        attrs.schemes.replace('\n', ' '), debugfs_schemes), shell=True,
        executable='/bin/bash')

def attr_str_ctx(damon_ctx):
    intervals = damon_ctx.intervals
    nr_regions = damon_ctx.nr_regions
    return '%d %d %d %d %d ' % (intervals.sample, intervals.aggr,
            intervals.ops_update, nr_regions.min_nr_regions,
            nr_regions.max_nr_regions)

def apply_kdamonds(kdamonds):
    if len(kdamonds) != 1:
        print('Currently only one kdamond is supported')
        exit(1)
    if len(kdamonds[0].contexts) != 1:
        print('currently only one damon_ctx is supported')
        exit(1)
    if len(kdamonds[0].contexts[0].targets) != 1:
        print('currently only one target is supported')
        exit(1)
    ctx = kdamonds[0].contexts[0]
    ret = subprocess.call('echo %s > %s' %
            (attr_str_ctx(ctx), debugfs_attrs), shell=True,
            executable='/bin/bash')
    if ret:
        return ret

    target = ctx.targets[0]
    if _damon.target_has_pid(ctx.ops):
        try:
            with open(debugfs_target_ids, 'w') as f:
                f.write('%d\n' % target.pid)
        except Exception as e:
            return e

        tid = target.pid
    else:
        try:
            with open(debugfs_target_ids, 'w') as f:
                f.write('paddr\n')
        except Exception as e:
            return e
        tid = 42
    if feature_supported('init_regions_target_idx'):
        tid = 0

    string = ' '.join(['%s %d %d' % (tid, r.start, r.end) for r in target.regions])
    return subprocess.call('echo "%s" > %s' % (string, debugfs_init_regions),
            shell=True, executable='/bin/bash')



#!/usr/bin/env python3

import argparse

parser = argparse.ArgumentParser()
parser.add_argument('--bar')
args = parser.parse_args()

if 'foo' in args:
    print('foo in args')
if 'bar' in args:
    print('bar in args')



#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0

"""
Transform binary trace data into human readable text that can be used for
heatmap drawing, or directly plot the data in a heatmap format.

Format of the text is:

    <time> <space> <heat>
    ...

"""

import argparse
import os
import subprocess
import sys
import tempfile

import _damon_result
import _damo_fmt_nr

class HeatPixel:
    time = None
    addr = None
    heat = None

    def __init__(self, time, addr, heat):
        self.time = time
        self.addr = addr
        self.heat = heat

def add_heats(snapshot, duration, pixels, time_unit, space_unit, addr_range):
    """Add heats in a monitoring 'snapshot' of specific time 'duration' to
    the corresponding heats 'pixels'.
    """
    pixel_sz = time_unit * space_unit

    for region in snapshot.regions:
        start = max(region.start, addr_range[0])
        end = min(region.end, addr_range[1])
        if start >= end:
            continue

        fraction_start = start
        addr_idx = int(float(fraction_start - addr_range[0]) / space_unit)
        while fraction_start < end:
            fraction_end = min((addr_idx + 1) * space_unit + addr_range[0],
                    end)
            heat = region.nr_accesses * duration * (
                    fraction_end - fraction_start)

            pixel = pixels[addr_idx]
            heat += pixel.heat * pixel_sz
            pixel.heat = float(heat) / pixel_sz

            fraction_start = fraction_end
            addr_idx += 1

def heat_pixels_from_snapshots(snapshots, time_range, addr_range, resols):
    """Get heat pixels for monitoring snapshots."""
    time_unit = (time_range[1] - time_range[0]) / float(resols[0])
    space_unit = (addr_range[1] - addr_range[0]) / float(resols[1])

    pixels = [[HeatPixel(int(time_range[0] + i * time_unit),
                    int(addr_range[0] + j * space_unit), 0.0)
            for j in range(resols[1])] for i in range(resols[0])]

    if len(snapshots) < 2:
        return pixels

    for idx, shot in enumerate(snapshots[1:]):
        start = shot.start_time
        end = min(shot.end_time, time_range[1])

        fraction_start = start
        time_idx = int(float(fraction_start - time_range[0]) / time_unit)
        while fraction_start < end:
            fraction_end = min((time_idx + 1) * time_unit + time_range[0], end)
            add_heats(shot, fraction_end - fraction_start, pixels[time_idx],
                    time_unit, space_unit, addr_range)
            fraction_start = fraction_end
            time_idx += 1
    return pixels

def heatmap_plot_ascii(pixels, time_range, addr_range, resols, colorset):
    highest_heat = None
    lowest_heat = None
    for snapshot in pixels:
        for pixel in snapshot:
            if highest_heat == None or highest_heat < pixel.heat:
                highest_heat = pixel.heat
            if lowest_heat == None or lowest_heat > pixel.heat:
                lowest_heat = pixel.heat
    if highest_heat == None and lowest_heat == None:
        return
    heat_unit = float(highest_heat + 1 - lowest_heat) / 9

    colorsets = {
        'gray':[
            [232] * 10,
            [237, 239, 241, 243, 245, 247, 249, 251, 253, 255]],
        'flame':[
            [232, 1, 1, 2, 3, 3, 20, 21,26, 27, 27],
            [239, 235, 237, 239, 243, 245, 247, 249, 251, 255]],
        'emotion':[
            [232, 234, 20, 21, 26, 2, 3, 1, 1, 1],
            [239, 235, 237, 239, 243, 245, 247, 249, 251, 255]],
        }
    colors = colorsets[colorset]
    for snapshot in pixels:
        chars = []
        for pixel in snapshot:
            heat = int(float(pixel.heat - lowest_heat) / heat_unit)
            heat = min(heat, len(colors[0]) - 1)
            bg = colors[0][heat]
            fg = colors[1][heat]
            chars.append(u'\u001b[48;5;%dm\u001b[38;5;%dm%d' %
                    (bg, fg, heat))
        print(''.join(chars) + u'\u001b[0m')
    color_samples = [u'\u001b[48;5;%dm\u001b[38;5;%dm %d ' %
            (colors[0][i], colors[1][i], i) for i in range(10)]
    print('# access_frequency: %s' % ''.join(color_samples) + u'\u001b[0m')
    print('# x-axis: space (%d-%d: %s)' % (addr_range[0], addr_range[1],
        _damo_fmt_nr.format_sz(addr_range[1] - addr_range[0], False)))
    print('# y-axis: time (%d-%d: %s)' % (time_range[0], time_range[1],
        _damo_fmt_nr.format_time(time_range[1] - time_range[0], False)))
    print('# resolution: %dx%d (%s and %s for each character)' % (
        len(pixels[1]), len(pixels),
        _damo_fmt_nr.format_sz(
            float(addr_range[1] - addr_range[0]) / len(pixels[1]), False),
        _damo_fmt_nr.format_time(
            float(time_range[1] - time_range[0]) / len(pixels), False)))

def pr_heats(args, damon_result):
    tid = args.tid
    tres = args.resol[0]
    tmin = args.time_range[0]
    tmax = args.time_range[1]
    ares = args.resol[1]
    amin = args.address_range[0]
    amax = args.address_range[1]

    tunit = (tmax - tmin) // tres
    aunit = (amax - amin) // ares

    # Compensate the values so that those fit with the resolution
    tmax = tmin + tunit * tres
    amax = amin + aunit * ares

    # __pr_heats(damon_result, tid, tunit, tmin, tmax, aunit, amin, amax)

    snapshots = damon_result.target_snapshots[tid]
    pixels = heat_pixels_from_snapshots(snapshots, [tmin, tmax], [amin, amax],
            [tres, ares])

    if args.heatmap == 'stdout':
        heatmap_plot_ascii(pixels, [tmin, tmax], [amin, amax], [tres, ares],
                args.stdout_heatmap_color)
        return

    for row in pixels:
        for pixel in row:
            time = pixel.time
            addr = pixel.addr
            if not args.abs_time:
                time -= tmin
            if not args.abs_addr:
                addr -= amin

            print('%s\t%s\t%s' % (time, addr, pixel.heat))

class GuideInfo:
    tid = None
    start_time = None
    end_time = None
    lowest_addr = None
    highest_addr = None
    gaps = None

    def __init__(self, tid, start_time):
        self.tid = tid
        self.start_time = start_time
        self.gaps = []

    def regions(self):
        regions = []
        region = [self.lowest_addr]
        for gap in self.gaps:
            for idx, point in enumerate(gap):
                if idx == 0:
                    region.append(point)
                    regions.append(region)
                else:
                    region = [point]
        region.append(self.highest_addr)
        regions.append(region)
        return regions

    def total_space(self):
        ret = 0
        for r in self.regions():
            ret += r[1] - r[0]
        return ret

    def __str__(self):
        lines = ['target_id:%d' % self.tid]
        lines.append('time: %d-%d (%s)' % (self.start_time, self.end_time,
                    _damo_fmt_nr.format_time(self.end_time - self.start_time,
                        False)))
        for idx, region in enumerate(self.regions()):
            lines.append('region\t%2d: %020d-%020d (%s)' %
                    (idx, region[0], region[1],
                        _damo_fmt_nr.format_sz(region[1] - region[0], False)))
        return '\n'.join(lines)

def is_overlap(region1, region2):
    if region1[1] < region2[0]:
        return False
    if region2[1] < region1[0]:
        return False
    return True

def overlap_region_of(region1, region2):
    return [max(region1[0], region2[0]), min(region1[1], region2[1])]

def overlapping_regions(regions1, regions2):
    overlap_regions = []
    for r1 in regions1:
        for r2 in regions2:
            if is_overlap(r1, r2):
                r1 = overlap_region_of(r1, r2)
        if r1:
            overlap_regions.append(r1)
    return overlap_regions

def get_guide_info(damon_result):
    "return the set of guide information for the moitoring result"
    guides = {}
    for snapshots in damon_result.target_snapshots.values():
        for snapshot in snapshots:
            monitor_time = snapshot.end_time
            tid = snapshot.target_id
            if not tid in guides:
                guides[tid] = GuideInfo(tid, monitor_time)
            guide = guides[tid]
            guide.end_time = monitor_time

            last_addr = None
            gaps = []
            for r in snapshot.regions:
                saddr = r.start
                eaddr = r.end

                if not guide.lowest_addr or saddr < guide.lowest_addr:
                    guide.lowest_addr = saddr
                if not guide.highest_addr or eaddr > guide.highest_addr:
                    guide.highest_addr = eaddr

                if not last_addr:
                    last_addr = eaddr
                    continue
                if last_addr != saddr:
                    gaps.append([last_addr, saddr])
                last_addr = eaddr

            if not guide.gaps:
                guide.gaps = gaps
            else:
                guide.gaps = overlapping_regions(guide.gaps, gaps)

    return sorted(list(guides.values()), key=lambda x: x.total_space(),
                    reverse=True)

def pr_guide(damon_result):
    for guide in get_guide_info(damon_result):
        print(guide)

def region_sort_key(region):
    return region[1] - region[0]

def set_missed_args(args, damon_result):
    if args.tid and args.time_range and args.address_range:
        return
    guides = get_guide_info(damon_result)
    guide = guides[0]
    if not args.tid:
        args.tid = guide.tid
    for g in guides:
        if g.tid == args.tid:
            guide = g
            break

    if not args.time_range:
        args.time_range = [guide.start_time, guide.end_time]

    if not args.address_range:
        args.address_range = sorted(guide.regions(), key=lambda x: x[1] - x[0],
                reverse=True)[0]

def plot_range(orig_range, use_absolute_val):
    plot_range = [x for x in orig_range]
    if not use_absolute_val:
        plot_range[0] -= orig_range[0]
        plot_range[1] -= orig_range[0]
    return plot_range

def plot_heatmap(data_file, output_file, args):
    terminal = output_file.split('.')[-1]
    if not terminal in ['pdf', 'jpeg', 'png', 'svg']:
        os.remove(data_file)
        print("Unsupported plot output type.")
        exit(-1)

    x_range = plot_range(args.time_range, args.abs_time)
    y_range = plot_range(args.address_range, args.abs_addr)

    gnuplot_cmd = """
    set term %s;
    set output '%s';
    set key off;
    set xrange [%f:%f];
    set yrange [%f:%f];
    set xlabel 'Time (ns)';
    set ylabel 'Address (bytes)';
    plot '%s' using 1:2:3 with image;""" % (terminal, output_file, x_range[0],
            x_range[1], y_range[0], y_range[1], data_file)
    subprocess.call(['gnuplot', '-e', gnuplot_cmd])
    os.remove(data_file)

def set_argparser(parser):
    parser.add_argument('--input', '-i', type=str, metavar='<file>',
            default='damon.data', help='input file name')
    parser.add_argument('--input_type', choices=['record', 'perf_script'],
            default=None, help='input file\'s type')

    parser.add_argument('--tid', metavar='<id>', type=int,
            help='target id')
    parser.add_argument('--resol', metavar='<resolution>', type=int, nargs=2,
            default=[500, 500],
            help='resolutions for time and address axises')
    parser.add_argument('--time_range', metavar='<time>', type=int, nargs=2,
            help='start and end time of the output')
    parser.add_argument('--address_range', metavar='<address>', type=int,
            nargs=2, help='start and end address of the output')
    parser.add_argument('--abs_time', action='store_true', default=False,
            help='display absolute time in output')
    parser.add_argument('--abs_addr', action='store_true', default=False,
            help='display absolute address in output')

    parser.add_argument('--guide', action='store_true',
            help='print a guidance for the ranges and resolution settings')
    parser.add_argument('--heatmap', metavar='<file>', type=str,
            help='heatmap image file to create.  stdout for terminal output')
    parser.add_argument('--stdout_heatmap_color',
            choices=['gray', 'flame', 'emotion'], default='gray',
            help='color theme for access frequencies')

def main(args=None):
    if not args:
        parser = argparse.ArgumentParser()
        set_argparser(parser)
        args = parser.parse_args()

    damon_result = _damon_result.parse_damon_result(
            args.input, args.input_type)
    if not damon_result:
        print('monitoring result file (%s) parsing failed' % args.input)
        exit(1)

    # Use 80x40 resolution as default for ascii plot
    if args.heatmap == 'stdout' and args.resol == [500, 500]:
        args.resol = [40, 80]

    if args.guide:
        pr_guide(damon_result)
    else:
        set_missed_args(args, damon_result)
        orig_stdout = sys.stdout
        if args.heatmap and args.heatmap != 'stdout':
            tmp_path = tempfile.mkstemp()[1]
            tmp_file = open(tmp_path, 'w')
            sys.stdout = tmp_file

        pr_heats(args, damon_result)

        if args.heatmap and args.heatmap != 'stdout':
            sys.stdout = orig_stdout
            tmp_file.flush()
            tmp_file.close()
            plot_heatmap(tmp_path, args.heatmap, args)

if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0

import _damon

def set_argparser(parser):
    parser.add_argument('type', choices=['supported', 'unsupported', 'all'],
            default='all', nargs='?',
            help='type of features to listed')
    _damon.set_common_argparser(parser)

def main(args=None):
    if not args:
        parser = argparse.ArgumentParser()
        set_argparser(parser)
        args = parser.parse_args()

    _damon.initialize(args)
    features = _damon.get_supported_features()

    for feature in sorted(features.keys()):
        if args.type == 'all':
            print('%s: %s' % (feature,
                'Supported' if features[feature] else 'Unsupported'))
        elif args.type == 'supported' and features[feature]:
            print(feature)
        elif args.type == 'unsupported' and not features[feature]:
            print(feature)

if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0

def format_sz(sz_bytes, machine_friendly):
    if machine_friendly:
        return '%d' % sz_bytes

    sz_bytes = float(sz_bytes)
    if sz_bytes > 1<<40:
        return '%.3f TiB' % (sz_bytes / (1<<40))
    if sz_bytes > 1<<30:
        return '%.3f GiB' % (sz_bytes / (1<<30))
    if sz_bytes > 1<<20:
        return '%.3f MiB' % (sz_bytes / (1<<20))
    if sz_bytes > 1<<10:
        return '%.3f KiB' % (sz_bytes / (1<<10))
    return '%d B' % sz_bytes

def format_time(time_ns, machine_friendly):
    if machine_friendly:
        return '%d' % time_ns

    time_ns = float(time_ns)
    if time_ns > 60000000000:
        return '%d m %.3f s' % (time_ns / 60000000000,
                (time_ns % 60000000000) / 1000000000)
    if time_ns > 1000000000:
        return '%.3f s' % (time_ns / 1000000000)
    if time_ns > 1000000:
        return '%.3f ms' % (time_ns / 1000000)
    if time_ns > 1000:
        return '%.3f us' % (time_ns / 1000)
    return '%d ns' % time_ns




#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0

"Control DAMON_LRU_SORT"

import argparse
import os
import time

import _damon

plrus_params_dir = '/sys/module/damon_lru_sort/parameters'
plrus_params = ['kdamond_pid', 'enabled', 'hot_thres_access_freq',
        'cold_min_age', 'quota_ms', 'quota_reset_interval_ms',
        'wmarks_interval', 'wmarks_high', 'wmarks_mid', 'wmarks_low',
        'sample_interval', 'aggr_interval', 'min_nr_regions', 'max_nr_regions',
        'monitor_region_start', 'monitor_region_end',
        'nr_lru_sort_tried_hot_regions', 'nr_lru_sorted_hot_regions',
        'bytes_lru_sort_tried_hot_regions', 'bytes_lru_sorted_hot_regions',
        'nr_lru_sort_tried_cold_regions', 'nr_lru_sorted_cold_regions',
        'bytes_lru_sort_tried_cold_regions', 'bytes_lru_sorted_cold_regions']

def chk_plrus_sysfs():
    if not os.path.isdir(plrus_params_dir):
        print('%s not found' % plrus_params_dir)
    for param in plrus_params:
        param_file = os.path.join(plrus_params_dir, param)
        if not os.path.isfile(param_file):
            print('%s file not found' % param_file)
            exit(1)

def set_param(param, val):
    if val == None:
        return
    path = os.path.join(plrus_params_dir, param)
    with open(path, 'w') as f:
        f.write('%s' % val)

def plrus_running():
    with open(os.path.join(plrus_params_dir, 'kdamond_pid')) as f:
        return f.read() != '-1\n'

def plrus_enable(on):
    if not on:
        set_param('enabled', 'N')
        while plrus_running():
            time.sleep(1)
        return

    set_param('enabled', 'N')
    while plrus_running():
        time.sleep(1)
    set_param('enabled', 'Y')
    while not plrus_running():
        time.sleep(1)
    return

def plrus_read_status():
    for param in plrus_params:
        param_file = os.path.join(plrus_params_dir, param)
        if not os.path.isfile(param_file):
            continue

        with open(param_file, 'r') as f:
            print('%s: %s' % (param, f.read().strip()))

def set_argparser(parser):
    parser.add_argument('action', type=str, nargs='?',
            choices=['status', 'enable', 'disable'], default='status',
            help='read status, enable, or disable DAMON_RECLAIM')
    parser.add_argument('--hot_thres_access_freq', type=int,
            metavar='<permil>',
            help='hot memory region access frequency threshold (permil)')
    parser.add_argument('--cold_min_age', type=int, metavar='<microseconds>',
            help='time threshold for cold memory regions identification (us)')
    parser.add_argument('--quota', type=int,
            metavar=('<ms>', '<ms>'), nargs=2,
            default=[None] * 2,
            help='time quota and quota reset interval in ms')
    parser.add_argument('--wmarks', type=int,
            metavar=('<us>', '<per-thousand>', '<per-thousand>',
                '<per-thousand>'),
            nargs=4, default=[None] * 4,
            help='watermarks check interval and three watermarks')
    parser.add_argument('--monitor_intervals', type=int,
            metavar='<microseconds>', nargs=2, default=[None] * 2,
            help='intervals for sampling and aggregation of DAMON')
    parser.add_argument('--nr_regions', type=int, metavar='<number>',
            nargs=2, default=[None] * 2,
            help='minimum and maximum number of DAMON memory regions')
    parser.add_argument('--monitor_region', type=int, metavar='<phy addr>',
            nargs=2, default=[None] * 2,
            help='start and end addresses of the target memory region')

def main(args=None):
    if not args:
        parser = argparse.ArgumentParser()
        set_argparser(parser)
        args = parser.parse_args()

    _damon.ensure_root_permission()
    chk_plrus_sysfs()

    if args.action == 'status':
        plrus_read_status()
        return

    set_param('hot_thres_access_freq', args.hot_thres_access_freq)
    set_param('cold_min_age', args.cold_min_age)
    set_param('quota_ms', args.quota[0])
    set_param('quota_reset_interval_ms', args.quota[1])
    set_param('wmarks_interval', args.wmarks[0])
    set_param('wmarks_high', args.wmarks[1])
    set_param('wmarks_mid', args.wmarks[2])
    set_param('wmarks_low', args.wmarks[3])
    set_param('sample_interval', args.monitor_intervals[0])
    set_param('aggr_interval', args.monitor_intervals[1])
    set_param('min_nr_regions', args.nr_regions[0])
    set_param('max_nr_regions', args.nr_regions[1])
    set_param('monitor_region_start', args.monitor_region[0])
    set_param('monitor_region_end', args.monitor_region[1])

    plrus_enable(args.action == 'enable')

if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0

"""
Contains core functions for DAMON sysfs control.
"""

import os
import time

import _damon

feature_supports = None

# Use only one kdamond, one context, and one target for now
root_dir = '/sys/kernel/mm/damon'
admin_dir = os.path.join(root_dir, 'admin')
kdamonds_dir = os.path.join(admin_dir, 'kdamonds')
kdamonds_nr_file = os.path.join(kdamonds_dir, 'nr_kdamonds')
kdamond_dir = os.path.join(kdamonds_dir, '0')
kdamond_state_file = os.path.join(kdamond_dir, 'state')
kdamond_pid_file = os.path.join(kdamond_dir, 'pid')
contexts_dir = os.path.join(kdamond_dir, 'contexts')
contexts_nr_file = os.path.join(contexts_dir, 'nr_contexts')
context_dir = os.path.join(contexts_dir, '0')
context_avail_operations_file = os.path.join(context_dir, 'avail_operations')
context_operations_file = os.path.join(context_dir, 'operations')
context_attrs_dir = os.path.join(context_dir, 'monitoring_attrs')
attrs_intervals_dir = os.path.join(context_attrs_dir, 'intervals')
intervals_sample_us_file = os.path.join(attrs_intervals_dir, 'sample_us')
intervals_aggr_us_file = os.path.join(attrs_intervals_dir, 'aggr_us')
intervals_update_us_file = os.path.join(attrs_intervals_dir, 'update_us')
attrs_nr_regions_dir = os.path.join(context_attrs_dir, 'nr_regions')
nr_regions_min_file = os.path.join(attrs_nr_regions_dir, 'min')
nr_regions_max_file = os.path.join(attrs_nr_regions_dir, 'max')
context_targets_dir = os.path.join(context_dir, 'targets')
targets_nr_file = os.path.join(context_targets_dir, 'nr_targets')
target_dir = os.path.join(context_targets_dir, '0')
target_pid_file = os.path.join(target_dir, 'pid_target')
target_regions_dir = os.path.join(target_dir, 'regions')
regions_nr_file = os.path.join(target_regions_dir, 'nr_regions')

def region_dir(region_idx):
    return os.path.join(target_regions_dir, '%d' % region_idx)

def region_start_file(region_idx):
    return os.path.join(region_dir(region_idx), 'start')

def region_end_file(region_idx):
    return os.path.join(region_dir(region_idx), 'end')

schemes_dir = os.path.join(context_dir, 'schemes')
schemes_nr_file = os.path.join(schemes_dir, 'nr_schemes')

def scheme_dir(scheme_idx):
    return os.path.join(schemes_dir, '%d' % scheme_idx)

# This class will be used in a future when we support multiple
# contexts/kdamonds
class DamonSysfsFile:
    indices = None  # e.g., {'kdamond': 0, 'context': 1}
    extra_path = None

    def __init__(self, indices, extra_path=None):
        self.indices = indices
        self.extra_path = extra_path

    def path(self):
        path = kdamonds_dir
        for keyword in ['kdamond', 'context', 'scheme', 'target',
                'region']:
            if keyword in self.indices:
                path = os.path.join(path, keyword + 's', self.indices[keyword])
        if self.extra_path:
            path = os.path.join(path, self.extra_path)
        return path

    def __str__(self):
        return self.path()

    def __repr__(self):
        return self.path()

    def regions_dir(self):
        return DamonSysfsFile(file_idx='regions',
                kdamond_idx=self.kdamond_idx, context_idx=self.context_idx,
                target_idx=self.target_idx)

    def regions_nr(self):
        return DamonSysfsFile(file_idx='regions/nr',
                kdamond_idx=self.kdamond_idx, context_idx=self.context_idx,
                target_idx=self.target_idx)

    def write(self, content):
        with open(self.path(), 'w') as f:
            f.write(content)

def _write(filepath, content):
    if _damon.pr_debug_log:
        print('write %s to %s' % (content, filepath))
    with open(filepath, 'w') as f:
        f.write(content)

def _read(filepath):
    with open(filepath, 'r') as f:
        return f.read()

def set_target(tid, init_regions):
    try:
        if tid == 'paddr':
            _write(context_operations_file, 'paddr')
        else:
            if init_regions:
                _write(context_operations_file, 'fvaddr')
            else:
                _write(context_operations_file, 'vaddr')
            _write(target_pid_file, '%d' % int(tid))

        _write(regions_nr_file, '%d' % len(init_regions))
        for idx, region in enumerate(init_regions):
            _write(region_start_file(idx), '%d' % region[0])
            _write(region_end_file(idx), '%d' % region[1])
        return 0
    except Exception as e:
        print(e)
        return 1

def turn_damon(on_off):
    if on_off == 'on':
        # In case of vaddr, too early monitoring shows unstable mapping changes.
        # Give the process a time to have stable memory mapping.
        time.sleep(0.5)
    try:
        _write(kdamond_state_file, on_off)
        return 0
    except Exception as e:
        print(e)
        return 1

def is_damon_running():
    return _read(kdamond_state_file).strip() == 'on'

def attrs_apply(attrs):
    if not attrs:
        return 0
    try:
        _write(intervals_sample_us_file, '%d' % attrs.sample_interval)
        _write(intervals_aggr_us_file, '%d' % attrs.aggr_interval)
        _write(intervals_update_us_file, '%d' % attrs.regions_update_interval)
        _write(nr_regions_min_file, '%d' % attrs.min_nr_regions)
        _write(nr_regions_max_file, '%d' % attrs.max_nr_regions)
        schemes = [x for x in attrs.schemes.split('\n') if x != '']
        _write(schemes_nr_file, '%d' % len(schemes))
        # access pattern
        for idx, scheme in enumerate(schemes):
            fields = scheme.split()
            field_idx = 0
            for pattern_dir in ['sz', 'nr_accesses', 'age']:
                for file_ in ['min', 'max']:
                    _write(os.path.join(scheme_dir(idx), 'access_pattern',
                        pattern_dir, file_), fields[field_idx])
                    field_idx += 1
            # action
            dbgfs_action_to_sysfs_action = {
                    0: 'willneed',
                    1: 'cold',
                    2: 'pageout',
                    3: 'hugepage',
                    4: 'nohugepage',
                    5: 'stat',
                    6: 'lru_prio',
                    7: 'lru_deprio',
                    }
            _write(os.path.join(scheme_dir(idx), 'action'),
                    dbgfs_action_to_sysfs_action[int(fields[field_idx])])
            field_idx += 1

            # quotas
            quotas_dir = os.path.join(scheme_dir(idx), 'quotas')
            for file_ in ['ms', 'bytes', 'reset_interval_ms']:
                _write(os.path.join(quotas_dir, file_), fields[field_idx])
                field_idx += 1
            weights_dir = os.path.join(quotas_dir, 'weights')
            for file_ in ['sz_permil', 'nr_accesses_permil', 'age_permil']:
                _write(os.path.join(weights_dir, file_), fields[field_idx])
                field_idx += 1

            # watermarks
            dbgfs_wmark_metroc_to_sysfs_metric = {
                    0: 'none',
                    1: 'free_mem_rate',
                    }
            wmarks_dir = os.path.join(scheme_dir(idx), 'watermarks')
            _write(os.path.join(wmarks_dir, 'metric'),
                    dbgfs_wmark_metroc_to_sysfs_metric[int(fields[field_idx])])
            field_idx += 1
            for file_ in ['interval_us', 'high', 'mid', 'low']:
                _write(os.path.join(wmarks_dir, file_), fields[field_idx])
                field_idx += 1
        return 0
    except Exception as e:
        print(e)
        return 1

# sysfs doesn't restore original state
def current_attrs():
    return None

def apply_kdamonds(kdamonds):
    if len(kdamonds) != 1:
        print('Currently only one kdamond is supported')
        exit(1)
    if len(kdamonds[0].contexts) != 1:
        print('currently only one damon_ctx is supported')
        exit(1)
    if len(kdamonds[0].contexts[0].targets) != 1:
        print('currently only one target is supported')
        exit(1)
    ensure_dirs_populated()
    ctx = kdamonds[0].contexts[0]
    try:
        _write(intervals_sample_us_file, '%d' % ctx.intervals.sample)
        _write(intervals_aggr_us_file, '%d' % ctx.intervals.aggr)
        _write(intervals_update_us_file, '%d' % ctx.intervals.ops_update)
        _write(nr_regions_min_file, '%d' % ctx.nr_regions.min_nr_regions)
        _write(nr_regions_max_file, '%d' % ctx.nr_regions.max_nr_regions)

        # TODO: Support schemes

        _write(context_operations_file, ctx.ops)

        target = ctx.targets[0]
        if _damon.target_has_pid(ctx.ops):
            _write(target_pid_file, '%d' % ctx.targets[0].pid)
        _write(regions_nr_file, '%d' % len(target.regions))
        for idx, region in enumerate(target.regions):
            _write(region_start_file(idx), '%d' % region.start)
            _write(region_end_file(idx), '%d' % region.end)
    except Exception as e:
        print(e)
        return 1

def commit_inputs():
    try:
        _write(kdamond_state_file, 'commit')
        return 0
    except Exception as e:
        print(e)
        return 1

def feature_supported(feature):
    if feature_supports == None:
        update_supported_features()
    return feature_supports[feature]

def get_supported_features():
    if feature_supports == None:
        update_supported_features()
    return feature_supports

populated = False
def ensure_dirs_populated():
    global populated
    if populated:
        return

    try:
        _write(kdamonds_nr_file, '1')
        _write(contexts_nr_file, '1')
        _write(targets_nr_file, '1')
        populated = True
    except Exception as e:
        print(e)
        print('failed populating kdamond and context dirs')
        exit(1)

def dirs_populated():
    return (int(_read(kdamonds_nr_file)) >= 1 and
            int(_read(contexts_nr_file)) >= 1 and
            int(_read(targets_nr_file)) >= 1)

def damon_sysfs_missed():
    'Return none-None if DAMON sysfs interface is not found'
    if not os.path.isdir(kdamonds_dir):
        return 'damon sysfs dir (%s) not found' % kdamonds_dir
    return None

def update_supported_features():
    global feature_supports

    if feature_supports != None:
        return None
    feature_supports = {x: False for x in _damon.features}

    missed = damon_sysfs_missed()
    if missed != None:
        return missed
    feature_supports = {x: True for x in _damon.features}
    feature_supports['record'] = False

    if not dirs_populated():
        ensure_dirs_populated()
    if not os.path.isfile(context_avail_operations_file):
        for feature in ['vaddr', 'paddr', 'fvaddr', 'vaddr']:
            try:
                _write(context_operations_file, feature)
                feature_supports[feature] = True
            except IOError:
                feature_supports[feature] = False
        return None

    avail_ops = _read(context_avail_operations_file).strip().split()
    for feature in ['vaddr', 'paddr', 'fvaddr']:
        feature_supports[feature] = feature in avail_ops

    return None

def initialize(args=None, skip_dirs_population=False):
    err = update_supported_features()
    if err:
        return err

    if not skip_dirs_population:
        ensure_dirs_populated()
    return None
